# -*- cperl -*- this is for emacs
#
# 
# Copyright (c) 2001-2002 Robert McLay
#
# Authors: Robert McLay <mclay@cfdlab.ae.utexas.edu>
#          Victor Parr  <vparr@mindspring.com>
#
# This file is part of PB: PerfBench.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

use strict;
my $VN = 'Routine Block Module $Revision: 1.13 $'."\n";

########################################################################
# RoutineBlk Class
########################################################################
package RoutineBlk;
sub new
{
  my $proto           = shift;
  my $class           = ref($proto) || $proto;
  my $self            = {};
  $self->{Name}     = shift;
  $self->{FlopType} = shift;
  $self->{Max}      = {};
  $self->{Min}      = {};
  $self->{Sum}      = {};
  $self->{np}       = 0;

  bless ($self, $class);
  return $self;
}

sub store
{
  my $self               = shift;
  my $d                  = shift;
  my $events             = shift;
  my $flag               = 1;
  $flag = shift if ($#_ > -1);

  my $max                = $self->{Max};
  my $min                = $self->{Min};
  my $sum                = $self->{Sum};
  my $np                 = $self->{np};
  my $flopType           = $self->{FlopType};
  my $nm;

  for $nm (keys %$events, "comm", "io")
    {
      my $key    = "${nm}_rate";
      my $ft     = (($nm eq "comm" || $nm eq "io") ? "wall" : $flopType);
      $d->{$key} = (($d->{$ft} > 0) ? $d->{$nm}/$d->{$ft} : 0);
      $d->{"${nm}_Ttl_rate"} = (($d->{"${ft}_Ttl"} > 0 && exists $d->{"${nm}_Ttl"}) ? 
                                $d->{"${nm}_Ttl"}/$d->{"${ft}_Ttl"} : 0);
    }

  for $nm (keys %$d)
    {
      my $value = $d->{$nm};

      $sum->{$nm}  = 0.0 if (! exists $sum->{$nm});
      $sum->{$nm} += $value;

      $min->{$nm}  = [$value, $np] if (! exists $min->{$nm} || $min->{$nm}[0] > $value);
      $max->{$nm}  = [$value, $np] if (! exists $max->{$nm} || $max->{$nm}[0] < $value);
    }

  $self->{np}++ if ($flag); 
}

sub average
{
  my $self      = shift;
  my $colName   = shift;
  my $inclusive = shift;
  my $flag      = ($#_ > -1) ? shift : "truncated";
  my $sum       = $self->{Sum};
  my $np        = $self->{np};
  my $ft        = ($inclusive) ? "$self->{FlopType}_Ttl"
                               :  $self->{FlopType};
  return $self->{Name} if ($colName eq "name");
  if (! exists $sum->{$colName})
    {
      die "Unknown column name: $colName";
    }

  my $value;
  my $name;
  if ( $colName =~ /(.*)_rate$/ )
    {
      $colName = $1;
      if (($colName eq "comm") || ($colName eq "io"))
        { $value = ($sum->{wall} > 0) ? $sum->{$colName}/$sum->{wall} : 0.0; }
      else
        { $value = ($sum->{$ft} > 0)  ? 1.0e-6*$sum->{$colName}/$sum->{$ft} : 0.0; }
    }
  else
    { $value = $sum->{$colName}/$np; }

  if ($flag eq "truncated")
    {
      $value = sprintf "%-12.5g", $value;
      $value =~ s/\s*$//;
      $value =~ s/^\s//;
    }
  return $value;
}

sub deltaPercent
{
  my $self    = shift;
  my $colName = shift;
  my $np      = $self->{np};
  my $sum     = $self->{Sum};
  my $max     = $self->{Max};
  my $min     = $self->{Min};
  return 0.0 if ($colName eq "name");
  die "Unknown column name: $colName" if (! exists $sum->{$colName});

  my $avg     = $sum->{$colName}/$np;
  return 0.0 if ($avg < 1.0e-4);
  

  my $value = ($max->{$colName}[0] - $min->{$colName}[0])*100/(2*$avg);
  return $value;
}

sub max
{
  my $self    = shift;
  my $colName = shift;
  my $flag    = "truncated";
  $flag       = shift if ($#_ > -1);
  my $max     = $self->{Max};
  return $self->{Name} if ($colName eq "name");
  die "Unknown column name: $colName"   if (! exists $max->{$colName});
  my $value = $max->{$colName}[0];
  if ($flag eq "truncated")
    {
      $value = sprintf "%-12.5g", $value;
      $value =~ s/\s*$//;
      $value =~ s/^\s//;
    }
  return $value;
}
  

sub min
{
  my $self    = shift;
  my $colName = shift;
  my $flag    = "truncated";
  $flag       = shift if ($#_ > -1);
  my $min     = $self->{Min};
  return $self->{Name} if ($colName eq "name");
  die "Unknown column name: $colName"   if (! exists $min->{$colName});
  my $value = $min->{$colName}[0];
  if ($flag eq "truncated")
    {
      $value = sprintf "%-12.5g", $value;
      $value =~ s/\s*$//;
      $value =~ s/^\s//;
    }
  return $value;
}

sub maxNP
{
  my $self    = shift;
  my $colName = shift;
  my $max     = $self->{Max};
  return $self->{Name} if ($colName eq "name");
  die "Unknown column name: $colName"   if (! exists $max->{$colName});
  return $max->{$colName}[1];
}

sub minNP
{
  my $self    = shift;
  my $colName = shift;
  my $min     = $self->{Min};
  return $self->{Name} if ($colName eq "name");
  die "Unknown column name: $colName"   if (! exists $min->{$colName});
  return $min->{$colName}[1];
}

sub np
{
  my $self    = shift;
  return $self->{np};
}

1;


