# -*- cperl -*- this is for emacs
#
# 
# Copyright (c) 2001-2002 Robert McLay
#
# Authors: Robert McLay <mclay@cfdlab.ae.utexas.edu>
#          Victor Parr  <vparr@mindspring.com>
#
# This file is part of PB: PerfBench.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

use strict;
my $VN = 'Table Control $Revision: 1.4 $'."\n";

########################################################################
# Tbl Class
########################################################################

package Tbl;
sub new
{
  my $proto           = shift;
  my $class           = ref($proto) || $proto;
  my $self            = {};
  my %args            = @_;
  $self->{columnStr}   = $args{COLUMNS};
  $self->{outputStyle} = $args{OUTPUTSTYLE};
  $self->{prtHdr}      = ((exists $args{PRTHDR})    ? $args{PRTHDR}    : 0);
  $self->{title}       = ((exists $args{TITLE})     ? $args{TITLE}     : "");
  $self->{inclusive}   = ((exists $args{INCLUSIVE}) ? $args{INCLUSIVE} : 0);
  $self->{events}      = ((exists $args{EVENTS})    ? $args{EVENTS}    : {});
  my @evlist           = (keys %{$args{EVENTS}});

  $self->{eventOrig}   = {};
  if ( exists $args{EVENTS})
    { 
      for (@evlist)
        {
          $self->{eventsOrig}->{$_} = 1;
        }
    }

  $self->{active}      = {};

  if ( exists $args{ACTIVE} )
    {
      my @actNames = split /:/, $args{ACTIVE};
      for (@actNames)
        {
          $self->{events}->{$_} = 1;
          $self->{active}->{$_} = 1;
        }
    }

  bless ($self, $class);
  return $self;
}

sub fields
{
  my $self      = shift;
  my $percentby = shift;
  my $ra        = shift;
  my $Ttl       = (($self->{inclusive}) ? "_Ttl" : "");

  my @cols      = split /:/, $self->{columnStr};

  my $i         = 0;
  for (@cols)
    {
      if ($_ eq "%PERCENTBY%")
        { $ra->[$i++] = "% $percentby" ; }
      elsif ($_ eq "%EVENTS%")
        {
          my $rh = $self->{events};
          my $key;
          for $key (keys %$rh)
            {
              $ra->[$i++] = "$key$Ttl";
              $ra->[$i++] = "$key${Ttl}_rate";
            }
        }
      else
        { $ra->[$i++] = $_; }
    }
}

sub active
{
  my $self      = shift;
  my $ra        = shift;
  @$ra          = (keys %{$self->{active}});
}

sub events
{
  my $self      = shift;
  my $ra        = shift;
  @$ra          = (keys %{$self->{eventsOrig}});
}

sub outputStyle
{
  my $self      = shift;
  return $self->{outputStyle};
}

sub getTitle
{
  my $self      = shift;
  return $self->{title};
}

sub prtHdr
{
  my $self      = shift;
  return $self->{prtHdr};
}

sub inclusive
{
  my $self      = shift;
  return $self->{inclusive};
}

1;


