# -*- cperl -*- this is for emacs
#
# 
# Copyright (c) 2001-2002 Robert McLay
#
# Authors: Robert McLay <mclay@cfdlab.ae.utexas.edu>
#          Victor Parr  <vparr@mindspring.com>
#
# This file is part of PB: PerfBench.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

use strict;
my $VN = 'Test Block Module $Revision: 1.18 $'."\n";

use RoutineBlk;
########################################################################
# TstBlk Class
########################################################################
package TstBlk;

use vars qw/ $overHeadCpu $overHeadWall %perf %FC_perf @eventNames/;
sub new
{
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my $self  = {};
  my $routines         = {};
  $self->{Events}      = {};
  $self->{FlopDir}     = "";
  $self->{HeadBlk}     = "";
  $self->{RoutineHash} = $routines;

  bless ($self, $class);
  return $self;
}

sub read
{
  my $self        = shift;
  my $filename    = shift;
  my $flopType    = shift;
  my $routines    = $self->{RoutineHash};
  my $events      = $self->{Events};
  my $name;

  undef %perf;
  undef @eventNames;

  eval $self->readPerlBlk($filename);
  die if $@;
  $self->{nEvents} = scalar(@eventNames);
  for $name (@eventNames)
    {
      $events->{$name} = 1; 
    }

  if ($self->{FlopDir})
    {
      my $flopDir =  $self->{FlopDir};
      my $bname   =  $filename;
      $bname      =~ s|.*/||;
      my $fname   =  "$flopDir/$bname";
      my $blk     = $self->readPerlBlk($fname);
      $blk        =~ s/^\$perf/\$FC_perf/gm;
      undef %FC_perf;
      eval $blk;
      die if $@;
      for $name (@eventNames)
        {
          $events->{$name} = 1; 
        }


      my @col;
      my $cnt = 0;
      for $name (@eventNames)
        {
          $col[$cnt++] = $name;
          $col[$cnt++] = "${name}_Ttl";
        }

      for $name (keys %perf)
        {
          next if ($name eq "/");
          my $colName;
          for $colName (@col)
            {
              if ((exists $FC_perf{$name}{$colName}) &&
                  ((! exists $perf{$name}{$colName}) || 
                   ($perf{$name}{$colName} == 0)))
                { $perf{$name}{$colName} = $FC_perf{$name}{$colName} }
            }
        }
    }

  my $first = 1;
  for $name (keys %perf)
    {
      next if ($name eq "/");
      $routines->{$name} = RoutineBlk->new($name,$flopType) if (! exists $routines->{$name});
      $routines->{$name}->store($perf{$name},$events);
      
      # Sum all routines in "/"
      my $Rname = "/";
      $routines->{$Rname} = RoutineBlk->new($Rname,$flopType) if (! exists $routines->{$Rname});
      $routines->{$Rname}->store($perf{$name},$events,$first);
      $first = 0 if ($first);
    }

}

sub readPerlBlk
{
  my $self        = shift;
  my $filename    = shift;
  my $columnNames = "";
  my $comment     = "";

  open(PB,$filename) || die "Can't open fileName: $filename: $!\n";
  while(<PB>)
    {
      $_ = remove_end_line_characters($_);
      if ($_ =~ /^\*\*\*/)
        { $comment .= $_."\n"; }
      elsif ($_ =~ /^ *$/)
        { last; }
    }
  $self->{HeadBlk} = $comment if ($self->{HeadBlk} eq "");

  # Skip to start of perl data block
  while (<PB>)
    {
      last if ($_ =~ /^### Start Perl Output ###/);
    }

  my $blk = "";
  while (<PB>)
    {
      last if ($_ =~ /^### End Perl Output ###/);
      $blk .= $_;
    }

  close(PB);
  die "Blank perl block" if ($blk eq "");
  return $blk;
}

sub output
{
  my $self        = shift;
  my ($tbl, $sortby, $percentby,$viewType,$percent) = @_;

  my $routines    = $self->{RoutineHash};
  my $outputStyle = $tbl->outputStyle();
  my $inclusive   = $tbl->inclusive();
  my ($name, $row, $column,$colName);
  my @table;

  my @fields;
  $tbl->fields($percentby, \@fields);

  print STDOUT "\n";

  print STDOUT "$self->{HeadBlk}\n" if ($tbl->prtHdr());

  $row    = 0;
  $column = 0;
  for (@fields)
    {
      $table[$row][$column++] = $_;
    }
  $row++;

  my $sort_hash = ($sortby eq "name") ? \&sort_hash_by_string 
                                      : \&sort_hash_by_number;
  $sortby    .="_Ttl" if ($inclusive);
  if ($sortby eq "io_Ttl" || $sortby eq "comm_Ttl")
    { $sortby = "wall_Ttl"; }
  elsif ($sortby =~ /(name|calls)_Ttl/)
    { $sortby = "$1"; }

  my %sortroutine;
  for $name (keys %$routines)
    {
      next if ($name eq "/");
      my $value = $routines->{$name}->average($sortby);
      $sortroutine{$name} = $value;
    }


  my $valueRoutine = "average";
  my $variationRoutine = "deltaPercent";
  if ($viewType eq "max")
    {
      $valueRoutine = "max";
      $variationRoutine = "maxNP";
    }
  elsif ($viewType eq "min")
    {
      $valueRoutine = "min";
      $variationRoutine = "minNP";
    }

  if ($outputStyle eq "user" || $outputStyle eq "spread")
    {
      my @active;
      my $max_cpu   = $routines->{"/"}->average("cpu");
      my $max_value = $routines->{"/"}->average($percentby);
      $percentby .="_Ttl" if ($inclusive);
      $tbl->active(\@active);
      my $printTable = 0;
      for $name (&$sort_hash(\%sortroutine))
        {
          my $cpu  = $routines->{$name}->average("cpu");
          next if ($cpu < $percent*$max_cpu);

          my $fnd = 0;
          for (@active)
            {
              $fnd = 1 if (abs($routines->{$name}->$valueRoutine($_,$inclusive)) > 1.e-20);
            }
          next if (scalar(@active) > 0 && ! $fnd);

          $printTable = 1;
          $column = 0;
          my $field;
          for $colName (@fields)
            {
              next if ($colName eq "name" || $colName =~ /^%/);
              if ($colName eq "np")
                {
                  $table[$row][$column++] = $routines->{$name}->np();
                  next;
                }
              $field  = $routines->{$name}->$valueRoutine($colName,$inclusive);
              my $delta  = $routines->{$name}->deltaPercent($colName);
              if ($delta > 0.01 && $outputStyle eq "user")
                {
                  if ($viewType eq "avg")
                    {
                      my $delta  = $routines->{$name}->deltaPercent($colName);
                      $delta  =  sprintf "%5.2f", $delta;
                      $delta  =~ s/^\s*//;
                      $field .=  " ($delta)";
                    }
                  else
                    {
                      my $delta  = $routines->{$name}->$variationRoutine($colName);
                      $field .=  " ($delta)";
                    }
                }
              $table[$row][$column++] = $field;
            }
          my $value         = $routines->{$name}->average($percentby);
          my $percent_value = ($max_value > 0.0) ? 100.0*$value/$max_value : 0.0;
          $percent_value    = ($percent_value > 100.0) ? 100.0 : $percent_value;
          $percent_value    = sprintf "%7.3f", $percent_value;
      
          $table[$row][$column++] = $percent_value;
          $table[$row][$column++] = $name;
          $row++;
        }
      
      if (! $printTable )
        { ; } 
      elsif ($outputStyle eq "user")
        {
          print STDOUT $tbl->getTitle()."\n";
          print STDOUT "Sorted by $sortby, Shown in $viewType mode:\n";
          beautify_table(\@table);
          
          for $row (0 .. $#table)
            {
              my $values = $table[$row];
              for $column ( 0 .. $#{$values})
                { print STDOUT "$values->[$column]   "; }
              print STDOUT "\n";
            }
        }
      else
        {
          for $row (0 .. $#table)
            {
              my $values = $table[$row];
              my $line   = "$values->[0]";
              for $column ( 1 .. $#{$values})
                { $line .= "\t$values->[$column]"; }
              print STDOUT "$line\n";
            }
        }
    }

  else
    {
      my @events;  $tbl->events(\@events);
      $row = 2;
      my $max_cpu   = $routines->{"/"}->average("cpu","full");
      my $max_value = $routines->{"/"}->average($percentby,"full");
      print STDOUT "\n### Start Perl Output ###\n";
      print STDOUT "\@eventNames=(";
      for (@events)
        { print STDOUT " \"$_\""; }
      print STDOUT " );\n";
      for $name (&$sort_hash(\%sortroutine))
        {
          my $cpu  = $routines->{$name}->average("cpu");
          next if $cpu < $percent*$max_cpu;
          my $np   = $routines->{$name}->np();
      
          print STDOUT "\$perf{\"$name\"}{np} = $np;\n";
      
          for $colName (@fields)
            {
              next if $colName eq "name" || $colName eq "np";
              my $field  = $routines->{$name}->average($colName,"full");
              print STDOUT "\$perf{\"$name\"}{$colName} = $field;\n";
            }
          my $value = $routines->{$name}->average($percentby,"full");
          my $percent_value = ($max_value > 0.0) ? 100.0*$value/$max_value : 0.0;
          $percent_value = sprintf "%11.5f", $percent_value;
          print STDOUT "\$perf{\"$name\"}{percent} = $percent_value;\n";
        }
      for $colName (@fields)
        {
          next if ($colName eq "name" || $colName eq "np");
          my $field  = $routines->{"/"}->average($colName,"full");
          print STDOUT "\$perf{\"/\"}{$colName} = $field;\n";
        }
      print STDOUT "### End Perl Output ###\n";
    }

  return;
}

sub events
{
  my $self = shift;
  return $self->{Events};
}

sub setFlopDir
{
  my $self         = shift;
  $self->{FlopDir} = shift;
}

sub remove_end_line_characters
{
    my($line) = @_;       # get line
    chomp($line);
    $_ = $line;
    s/\r$//;
    return $_;
}

sub sort_hash_by_number
{
  my $x     = shift;
  my %array = %$x;
  sort { $array{$b} <=> $array{$a} } keys %array;
}

sub sort_hash_by_string
{
  my $x     = shift;
  my %array = %$x;
  sort { $array{$a} cmp $array{$b}; } keys %array;
}

sub beautify_table
{

  # Takes a list-of-lists of text and beautifies it for printing
  # NOTES: * List arguments must be passed using explicit pass-by-reference, that is
  #          beautify_table(\@list_one, \@list_two)
  #        * Can optionally take two lists, where the second list is
  #          used for calculating the spacing used to print the first

  my($LoL, $sizeLoL);
  my(@max_col_size);
  my($i, $j);
  my $fillstring = "";
  $LoL     = \@{$_[0]};
  if (defined @{$_[1]}) {
    $sizeLoL = \@{$_[1]};
  } else {
    $sizeLoL = \@{$_[0]};
  }
  
  # Figure out max size in each column
  for $j ( 0 .. $#{@{${$sizeLoL}[0]}} - 1 ) {       # Loop over cols in row 0 to initialize
    $max_col_size[$j] = length(${$sizeLoL}[0][$j]);
  }
  for $i ( 1 .. $#{$sizeLoL} ) {                    # Loop over rows
    for $j ( 0 .. $#{@{${$sizeLoL}[$i]}} - 1 ) {    # Loop over cols in row
      $max_col_size[$j] = op_max($max_col_size[$j],length(${$sizeLoL}[$i][$j]));
    }
  }
  
  # Now add spacing
  for $i ( 0 .. $#{$LoL} ) {                        # Loop over rows
    for $j ( 0 .. $#{@{${$LoL}[$i]}} - 1 ) {        # Loop over cols in row
      foreach ( length($fillstring) .. ($max_col_size[$j] - 1) ) 
        { $fillstring .= " " };
      ${$LoL}[$i][$j] = "${$LoL}[$i][$j]".substr($fillstring, 0,($max_col_size[$j] - length(${$LoL}[$i][$j])));
    }
  }

}

sub op_max
{
  # Finds max in a list using ">" operator
  #
  my @values = @_;
  my $max        ;
  
  $max = $values[0];
  
  foreach (@values)
    {
      if ($_ > $max)
        {
          $max = $_;
        }
    }
  
  return $max;
}

1;


