/*
 * 
 * Copyright (c) 2001-2002 Robert McLay
 *
 * Authors: Robert McLay <mclay@cfdlab.ae.utexas.edu>
 *          Victor Parr  <vparr@mindspring.com>
 */

/*
 * This file is part of PB: PerfBench.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include "pb.h"
#include "events.h"
#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_LIBPAPI
#  include <papi.h>
#  include <papiStdEventDefs.h>
#endif

static int    eventList[MAX_EVENTS];
static Events events[MAX_EVENTS];

#define VN "$Revision: 1.7 $\n"

/* These list of events come from the file: papiStdEventDefs.h
   
   As it says there not all these events work on all platforms.
   They can also mean different things on different machines as well.
*/

#ifdef HAVE_LIBPAPI
static Events palete[]=
{
  {"PAPI_FP_INS",  "flops",   PAPI_FP_INS},    /*Floating point instructions executed*/
  {"PAPI_L1_DCM",  "L1_DCM",  PAPI_L1_DCM},    /*Level 1 data cache misses*/
  {"PAPI_L2_DCM",  "L2_DCM",  PAPI_L2_DCM},    /*Level 2 data cache misses*/
  {"PAPI_L2_ICM",  "L2_ICM",  PAPI_L2_ICM},    /*Level 2 instruction cache misses*/ 
  {"PAPI_L3_DCM",  "L3_DCM",  PAPI_L3_DCM},    /*Level 3 data cache misses*/
  {"PAPI_L3_ICM",  "L3_ICM",  PAPI_L3_ICM},    /*Level 3 instruction cache misses*/
  {"PAPI_L1_TCM",  "L1_TCM",  PAPI_L1_TCM},    /*Level 1 total cache misses*/
  {"PAPI_L2_TCM",  "L2_TCM",  PAPI_L2_TCM},    /*Level 2 total cache misses*/
  {"PAPI_L3_TCM",  "L3_TCM",  PAPI_L3_TCM},    /*Level 3 total cache misses*/
  {"PAPI_CA_SNP",  "CA_SNP",  PAPI_CA_SNP},    /*Snoops*/
  {"PAPI_CA_SHR",  "CA_SHR",  PAPI_CA_SHR},    /*Request for shared cache line (SMP)*/
  {"PAPI_CA_CLN",  "CA_CLN",  PAPI_CA_CLN},    /*Request for clean cache line (SMP)*/
  {"PAPI_CA_INV",  "CA_INV",  PAPI_CA_INV},    /*Request for cache line Invalidation (SMP)*/
  {"PAPI_CA_ITV",  "CA_ITV",  PAPI_CA_ITV},    /*Request for cache line Intervention (SMP)*/
  {"PAPI_L3_LDM",  "L3_LDM",  PAPI_L3_LDM},    /*Level 3 load misses */
  {"PAPI_L3_STM",  "L3_STM",  PAPI_L3_STM},    /*Level 3 store misses */
  {"PAPI_BRU_IDL", "BRU_IDL", PAPI_BRU_IDL},   /*Cycles branch units are idle*/
  {"PAPI_FXU_IDL", "FXU_IDL", PAPI_FXU_IDL},   /*Cycles integer units are idle*/
  {"PAPI_FPU_IDL", "FPU_IDL", PAPI_FPU_IDL},   /*Cycles floating point units are idle*/
  {"PAPI_LSU_IDL", "LSU_IDL", PAPI_LSU_IDL},   /*Cycles load/store units are idle*/
  {"PAPI_TLB_DM",  "TLB_DM",  PAPI_TLB_DM},    /*Data translation lookaside buffer misses*/
  {"PAPI_TLB_IM",  "TLB_IM",  PAPI_TLB_IM},    /*Instr translation lookaside buffer misses*/
  {"PAPI_TLB_TL",  "TLB_TL",  PAPI_TLB_TL},    /*Total translation lookaside buffer misses*/
  {"PAPI_L1_LDM",  "L1_LDM",  PAPI_L1_LDM},    /*Level 1 load misses */
  {"PAPI_L1_STM",  "L1_STM",  PAPI_L1_STM},    /*Level 1 store misses */
  {"PAPI_L2_LDM",  "L2_LDM",  PAPI_L2_LDM},    /*Level 2 load misses */
  {"PAPI_L2_STM",  "L2_STM",  PAPI_L2_STM},    /*Level 2 store misses */
  {"PAPI_BTAC_M",  "BTAC_M",  PAPI_BTAC_M},    /*BTAC miss*/
  {"PAPI_PRF_DM",  "PRF_DM",  PAPI_PRF_DM},    /*Prefetch data instruction caused a miss */
  {"PAPI_L3_DCH",  "L3_DCH",  PAPI_L3_DCH},    /*Level 3 Data Cache Hit*/
  {"PAPI_TLB_SD",  "TLB_SD",  PAPI_TLB_SD},    /*Xlation lookaside buffer shootdowns (SMP)*/
  {"PAPI_CSR_FAL", "CSR_FAL", PAPI_CSR_FAL},   /*Failed store conditional instructions*/
  {"PAPI_CSR_SUC", "CSR_SUC", PAPI_CSR_SUC},   /*Successful store conditional instructions*/
  {"PAPI_CSR_TOT", "CSR_TOT", PAPI_CSR_TOT},   /*Total store conditional instructions*/
  {"PAPI_MEM_SCY", "MEM_SCY", PAPI_MEM_SCY},   /*Cycles Stalled Waiting for Memory Access*/
  {"PAPI_MEM_RCY", "MEM_RCY", PAPI_MEM_RCY},   /*Cycles Stalled Waiting for Memory Read*/
  {"PAPI_MEM_WCY", "MEM_WCY", PAPI_MEM_WCY},   /*Cycles Stalled Waiting for Memory Write*/
  {"PAPI_STL_ICY", "STL_ICY", PAPI_STL_ICY},   /*Cycles with No Instruction Issue*/
  {"PAPI_FUL_ICY", "FUL_ICY", PAPI_FUL_ICY},   /*Cycles with Maximum Instruction Issue*/
  {"PAPI_STL_CCY", "STL_CCY", PAPI_STL_CCY},   /*Cycles with No Instruction Completion*/
  {"PAPI_FUL_CCY", "FUL_CCY", PAPI_FUL_CCY},   /*Cycles with Maximum Instruction Completion*/
  {"PAPI_HW_INT",  "HW_INT",  PAPI_HW_INT},    /*Hardware interrupts */
  {"PAPI_BR_UCN",  "BR_UCN",  PAPI_BR_UCN},    /*Unconditional branch instructions executed*/
  {"PAPI_BR_CN",   "BR_CN",   PAPI_BR_CN},     /*Conditional branch instructions executed*/
  {"PAPI_BR_TKN",  "BR_TKN",  PAPI_BR_TKN},    /*Conditional branch instructions taken*/
  {"PAPI_BR_NTK",  "BR_NTK",  PAPI_BR_NTK},    /*Conditional branch instructions not taken*/
  {"PAPI_BR_MSP",  "BR_MSP",  PAPI_BR_MSP},    /*Conditional branch instructions mispred*/
  {"PAPI_BR_PRC",  "BR_PRC",  PAPI_BR_PRC},    /*Conditional branch instructions corr. pred*/
  {"PAPI_FMA_INS", "FMA_INS", PAPI_FMA_INS},   /*FMA instructions completed*/
  {"PAPI_TOT_IIS", "TOT_IIS", PAPI_TOT_IIS},   /*Total instructions issued*/
  {"PAPI_TOT_INS", "TOT_INS", PAPI_TOT_INS},   /*Total instructions executed*/
  {"PAPI_INT_INS", "INT_INS", PAPI_INT_INS},   /*Integer instructions executed*/
  {"PAPI_LD_INS",  "LD_INS",  PAPI_LD_INS},    /*Load instructions executed*/
  {"PAPI_SR_INS",  "SR_INS",  PAPI_SR_INS},    /*Store instructions executed*/
  {"PAPI_BR_INS",  "BR_INS",  PAPI_BR_INS},    /*Total branch instructions executed*/
  {"PAPI_VEC_INS", "VEC_INS", PAPI_VEC_INS},   /*Vector/SIMD instructions executed*/
  {"PAPI_RES_STL", "RES_STL", PAPI_RES_STL},   /*Cycles processor is stalled on resource*/
  {"PAPI_FP_STAL", "FP_STAL", PAPI_FP_STAL},   /*Cycles any FP units are stalled */
  {"PAPI_TOT_CYC", "TOT_CYC", PAPI_TOT_CYC},   /*Total cycles*/
  {"PAPI_IPS",     "IPS",     PAPI_IPS},       /*Instructions executed per second*/
  {"PAPI_LST_INS", "LST_INS", PAPI_LST_INS},   /*Total load/store inst. executed*/
  {"PAPI_SYC_INS", "SYC_INS", PAPI_SYC_INS},   /*Sync. inst. executed */
  {"PAPI_L1_DCH",  "L1_DCH",  PAPI_L1_DCH},    /*L1 D Cache Hit*/
  {"PAPI_L2_DCH",  "L2_DCH",  PAPI_L2_DCH},    /*L2 D Cache Hit*/
  {"PAPI_L1_DCA",  "L1_DCA",  PAPI_L1_DCA},    /*L1 D Cache Access*/
  {"PAPI_L2_DCA",  "L2_DCA",  PAPI_L2_DCA},    /*L2 D Cache Access*/
  {"PAPI_L3_DCA",  "L3_DCA",  PAPI_L3_DCA},    /*L3 D Cache Access*/
  {"PAPI_L1_DCR",  "L1_DCR",  PAPI_L1_DCR},    /*L1 D Cache Read*/
  {"PAPI_L2_DCR",  "L2_DCR",  PAPI_L2_DCR},    /*L2 D Cache Read*/
  {"PAPI_L3_DCR",  "L3_DCR",  PAPI_L3_DCR},    /*L3 D Cache Read*/
  {"PAPI_L1_DCW",  "L1_DCW",  PAPI_L1_DCW},    /*L1 D Cache Write*/
  {"PAPI_L2_DCW",  "L2_DCW",  PAPI_L2_DCW},    /*L2 D Cache Write*/
  {"PAPI_L3_DCW",  "L3_DCW",  PAPI_L3_DCW},    /*L3 D Cache Write*/
  {"PAPI_L1_ICH",  "L1_ICH",  PAPI_L1_ICH},    /*L1 instruction cache hits*/
  {"PAPI_L2_ICH",  "L2_ICH",  PAPI_L2_ICH},    /*L2 instruction cache hits*/
  {"PAPI_L3_ICH",  "L3_ICH",  PAPI_L3_ICH},    /*L3 instruction cache hits*/
  {"PAPI_L1_ICA",  "L1_ICA",  PAPI_L1_ICA},    /*L1 instruction cache accesses*/
  {"PAPI_L2_ICA",  "L2_ICA",  PAPI_L2_ICA},    /*L2 instruction cache accesses*/
  {"PAPI_L3_ICA",  "L3_ICA",  PAPI_L3_ICA},    /*L3 instruction cache accesses*/
  {"PAPI_L1_ICR",  "L1_ICR",  PAPI_L1_ICR},    /*L1 instruction cache reads*/
  {"PAPI_L2_ICR",  "L2_ICR",  PAPI_L2_ICR},    /*L2 instruction cache reads*/
  {"PAPI_L3_ICR",  "L3_ICR",  PAPI_L3_ICR},    /*L3 instruction cache reads*/
  {"PAPI_L1_ICW",  "L1_ICW",  PAPI_L1_ICW},    /*L1 instruction cache writes*/
  {"PAPI_L2_ICW",  "L2_ICW",  PAPI_L2_ICW},    /*L2 instruction cache writes*/
  {"PAPI_L3_ICW",  "L3_ICW",  PAPI_L3_ICW},    /*L3 instruction cache writes*/
  {"PAPI_L1_TCH",  "L1_TCH",  PAPI_L1_TCH},    /*L1 total cache hits*/
  {"PAPI_L2_TCH",  "L2_TCH",  PAPI_L2_TCH},    /*L2 total cache hits*/
  {"PAPI_L3_TCH",  "L3_TCH",  PAPI_L3_TCH},    /*L3 total cache hits*/
  {"PAPI_L1_TCA",  "L1_TCA",  PAPI_L1_TCA},    /*L1 total cache accesses*/
  {"PAPI_L2_TCA",  "L2_TCA",  PAPI_L2_TCA},    /*L2 total cache accesses*/
  {"PAPI_L3_TCA",  "L3_TCA",  PAPI_L3_TCA},    /*L3 total cache accesses*/
  {"PAPI_L1_TCR",  "L1_TCR",  PAPI_L1_TCR},    /*L1 total cache reads*/
  {"PAPI_L2_TCR",  "L2_TCR",  PAPI_L2_TCR},    /*L2 total cache reads*/
  {"PAPI_L3_TCR",  "L3_TCR",  PAPI_L3_TCR},    /*L3 total cache reads*/
  {"PAPI_L1_TCW",  "L1_TCW",  PAPI_L1_TCW},    /*L1 total cache writes*/
  {"PAPI_L2_TCW",  "L2_TCW",  PAPI_L2_TCW},    /*L2 total cache writes*/
  {"PAPI_L3_TCW",  "L3_TCW",  PAPI_L3_TCW},    /*L3 total cache writes*/
  {"PAPI_FML_INS", "FML_INS", PAPI_FML_INS},   /*FM ins */
  {"PAPI_FAD_INS", "FAD_INS", PAPI_FAD_INS},   /*FA ins */
  {"PAPI_FDV_INS", "FDV_INS", PAPI_FDV_INS},   /*FD ins */
  {"PAPI_FSQ_INS", "FSQ_INS", PAPI_FSQ_INS},   /*FSq ins */
  {"PAPI_FNV_INS", "FNV_INS", PAPI_FNV_INS},   /*Finv ins */
  {0             ,  0       , 0}  
};
#endif

int parseStr(const char* str, int lenStr, int** eventListIn, Events** eventsIn)
{
#ifdef HAVE_LIBPAPI
  Events* tabptr;
  int i,len;
  int nEvent = 0;
  const char * p = str;
  while (1)
    {
      const char* start = p;
      for (i = 0; i < lenStr; i++, p++)
        if (*p == ':') break;
      len = p - start;
      p++;
          
      if (lenStr <= 0) break;
      lenStr -= (len+1);

      for (tabptr = &palete[0]; tabptr->papi_name; tabptr++)
        {
          if (tabptr->papi_name[len] == 0 &&
              STRNCASECMP(start, ==, tabptr->papi_name,len) )
            {
              if (nEvent >= MAX_EVENTS)
                { 
                  fprintf(stderr,"Too many events requested, increase MAX_EVENTS\n");
                  exit(1);
                }
              events[nEvent].papi_name = tabptr->papi_name;
              events[nEvent].pb_name   = tabptr->pb_name;
              events[nEvent].eventNum  = tabptr->eventNum;
              eventList[nEvent]        = tabptr->eventNum;
              nEvent++;
              break;
            }
        }
    }

  *eventListIn = &eventList[0];
  *eventsIn    = &events[0];
  return nEvent;
#else
  return 0;
#endif
}



