/*
 * This file is part of PB: PerfBench.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include "cpb.h"

static char *reserved[] = 
{
  "auto",
  "break",
  "case",
  "char",
  "const",
  "continue",
  "default",
  "do",
  "double",
  "else",
  "enum",
  "extern",
  "float",
  "for",
  "goto",
  "if",
  "int",
  "long",
  "register",
  "return",
  "short",
  "signed",
  "sizeof",
  "static",
  "struct",
  "switch",
  "typedef",
  "union",
  "unsigned",
  "void",
  "volatile",
  "while",
  "exit",  /* this is not a reserved word, need it for PBend */
  0
};

int lineno = 0;

Token gettoken(char* word, FILE* fp) 
{
  enum {NEUTRAL, INQUOTE, INIDENT, INWORD, BCOMMENT, INCOMMENT, ECOMMENT}
  state = NEUTRAL;
  int c;
  char *w;

  w = word;
  while ((c = fgetc(fp)) != EOF) 
    {
      if (c == '\n')
	lineno++;
      switch (state) {
      case NEUTRAL:
	switch (c) 
	  {
	    /*
	      case '#':
	      return THASH;
	    */
	  case ',':
	    return TCOMMA;
	  case ';':
	    return TSEMI;
	  case '(':
	    return TLPAREN;
	  case ')':
	    return TRPAREN;
	  case '{':
	    return TLBRACE;
	  case '}':
	    return TRBRACE;
	  case '*':
	    return TSTAR;
	  case '\n':
	    return TNL;
	  case ' ':
	    return TSPACE;
	  case '\t':
	    return TTAB;
	  case '"':
	    state = INQUOTE;
	    *w++ = c;
	    continue;
	  case '/':
	    state = BCOMMENT;
	    continue;
	  default:
	    if (isalpha(c)) {
	      *w++ = c;
	      state = INIDENT;
	      continue;
	    }
	    else if (isalnum(c) || c == '#') {
	      *w++ = c;
	      state = INWORD;
	      continue;
	    }
	    *w++ = c;
	    *w = '\0';
	    return TOTHER;
	  }
      case INQUOTE:
	*w++ = c;
	switch (c) {
	case '\\':
	  *w++ = fgetc(fp);
	  continue;
	case '"':
	  *w = '\0';
	  /*fprintf(stderr, "INQ: %s\n", word);*/
	  return TQUOTE;
	default:
	  continue;
	}
      case BCOMMENT:
	if (c == '*') {
	  state = INCOMMENT;
	  *w++ = '/';
	  *w++ = '*';
	  continue;
	}
	ungetc(c, fp);
	return TSLASH;
      case INCOMMENT:
	*w++ = c;
	if (c == '*')
	  state = ECOMMENT;
	continue;
      case ECOMMENT:
	*w++ = c;
	if (c == '/') {
	  *w = '\0';
	  return TCOMMENT;
	}
	if (c != '*')
	  state = INCOMMENT;
	continue;
      case INIDENT:
	if (isalnum(c) || c == '_') {
	  *w++ = c;
	  continue;
	}
	ungetc(c, fp);
	*w = '\0';
	{ char **p = 0;
	for (p = reserved; *p; *p++)
	  if (strcmp(*p, word) == 0)
	    return TRESERVED;
	}
	return TIDENT;
      case INWORD:
	/*
	  switch (c) {
	  case ' ':
	  case '\t':
	  case '\n':
	  case '/':
	  case ';':
	  case ',':
	  ungetc(c, fp);
	  *w = '\0';
	  fprintf(stderr, "TWORD: %s\n", word);
	  return TWORD;
	  default:
	  *w++ = c;
	  continue;
	  }
	*/
	if (!(isalnum(c))) {
	  ungetc(c, fp);
	  *w = '\0';
	  return TWORD;
	}
	*w++ = c;
	continue;
      }
    }
  return TEOF;
}
