/*
 * 
 * Copyright (c) 2001-2002 Robert McLay
 *
 * Authors: Robert McLay <mclay@cfdlab.ae.utexas.edu>
 *          Victor Parr  <vparr@mindspring.com>
 */

/*
 * This file is part of PB: PerfBench.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef HASH_H
#define HASH_H

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#define FREE_ADT 1 
#define KEEP_ADT 0

#define HASH_TABLE_SIZE(ht) ht->num_entries

typedef struct Link_struct Link;
struct Link_struct
{
  Link*    next;
  char*    key;
  void*    adt;
};

typedef struct Bucket_struct
{
  int      num_entries;
  Link*    link;
} Buckets;

typedef struct Hash_struct 
{
  int          num_buckets;
  int          num_entries;
  Buckets*     buckets;
  unsigned int (*hash_fn)(const char *);
  int          (*match_fn)(const char * a, const char * b);
  size_t       (*length_fn)(const char *a);
} Hash;


Hash* make_hashtable(int size, 
                     unsigned int (*hash_fn)(const char *),
                     int (*match_fn)(const char *, const char *),
                     size_t (*length_fn)(const char *));

void free_hashtable(Hash* ht, int flag);

void* insert_in_hash(Hash* ht, char* key, void* adt, void* result);

void* find_in_hash(Hash* ht, char* key);

void for_all_in_hash(Hash* ht, void (*func)(const char *,void *));

#if (defined(NEED_PROTO_STRCASECMP) || !defined(HAVE_STRCASECMP))
  int strcasecmp(const char * s1, const char * s2);
#endif

unsigned int hashUC(const char * string);


#endif


