/*
 * 
 * Copyright (c) 2001-2002 Robert McLay
 *
 * Authors: Robert McLay <mclay@cfdlab.ae.utexas.edu>
 *          Victor Parr  <vparr@mindspring.com>
 */

/*
 * This file is part of PB: PerfBench.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include "pb.h"
#include "events.h"
#ifdef HAVE_LIBPAPI
#  include <papi.h>
#endif
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/times.h>
#include <stdio.h>

double        SecPerClk;
static double tckclk    = 0.0;
#define VN "measure.c: $Revision: 1.11 $ $Date: 2002/05/22 02:24:41 $\n"


static int   PapiEventSet;
static int   HdwrFlopCntr = 0;

double tckclk_init(void)
{
  long clktck = 0;

  if ( (clktck = sysconf(_SC_CLK_TCK)) < 0)
    exit(1);

  tckclk = 1.0/((double) clktck);
  return tckclk;
}

#if (defined(_CRAY) && !defined(_CRAYMPP))
float cputime(float *t)
{
  float t1;
  clock_t result;
  result = clock();
  t1 = ((float) result)/((float) CLOCKS_PER_SEC);
  *t = t1;
  return t1;
}
#else
double cputime(double * t)
{
  struct tms use;

  (void) times(&use);
  *t = ((double) (use.tms_utime+use.tms_stime))*tckclk;
  return (*t);
}
#endif

#if (defined(_CRAY) && !defined(_CRAYMPP))
float walltime(float *t)
{
  float t1;
  clock_t result;
  result = clock();
  t1 = ((float) result)/((float) CLOCKS_PER_SEC);
  *t = t1;
  return t1;
}
#else
double walltime(double * t)
{
  double t1;
  struct timezone z;
  struct timeval  tv;
  gettimeofday(&tv, &z);
  t1 = tv.tv_sec + tv.tv_usec * 1.0e-6;
  *t = t1;
  return t1;
}
#endif

int papi_error(int code, char *str)
{
#ifdef HAVE_LIBPAPI  
  char out[PAPI_MAX_STR_LEN];
  PAPI_perror(code, out, PAPI_MAX_STR_LEN);
  fprintf(stderr,"Error in %s: %s\n",str,out);
#endif
  return code;
}

int counter_init(int nEvents, int* eventList)
{
#ifdef HAVE_LIBPAPI  
  int retval;
  INT64 tmp[MAX_EVENTS];

  if (nEvents < 1) return;
  /* Initialize */
  PapiEventSet = PAPI_NULL;
  if ((retval = PAPI_library_init(PAPI_VER_CURRENT)) != PAPI_VER_CURRENT)
    {
      fprintf(stderr,"PAPI_library_init(PAPI_VER_CURRENT): %d, PAPI_VER_CURRENT: %d\n",
              retval, PAPI_VER_CURRENT);
      return papi_error(retval, "PAPI_library_init(PAPI_VER_CURRENT)");
    }
  
  if ((retval = PAPI_set_debug(PAPI_VERB_ECONT)) != PAPI_OK)
    return papi_error(retval, "PAPI_set_debug(PAPI_VERB_ECONT)");
  
  /* Create the EventSet */
  if ((retval = PAPI_create_eventset(&PapiEventSet)) != PAPI_OK)
    return papi_error(retval, "PAPI_create_eventset(&EventSet)");

  /* Add the FP count event to our EventSet */
  if ((retval = PAPI_add_events(&PapiEventSet, eventList, nEvents)) != PAPI_OK)
    return papi_error(retval, "PAPI_add_events(&PapiEventSet, eventList, nEvents)");

  /* Check for Start/Read with no errors. */
  if ((retval = PAPI_start(PapiEventSet)) != PAPI_OK)
    return papi_error(retval, "PAPI_start(EventSet)");
  if ((retval = PAPI_read(PapiEventSet, &tmp[0])) != PAPI_OK)
    return papi_error(retval, "PAPI_read(EventSet, &tmp)");

  HdwrFlopCntr = 1;     /* TRUE when hardware flop counting */
#endif
  return 0;
}

void counter_read(INT64* cnt)
{
  int retval;
#ifdef HAVE_LIBPAPI  
  if (HdwrFlopCntr)
    {
      retval = PAPI_read(PapiEventSet, cnt);
      /* assert(retval == PAPI_OK); */
    }
#else
  *cnt = 0;
#endif
}



