/*
 * 
 * Copyright (c) 2001-2002 Robert McLay
 *
 * Authors: Robert McLay <mclay@cfdlab.ae.utexas.edu>
 *          Victor Parr  <vparr@mindspring.com>
 */

/*
 * This file is part of PB: PerfBench.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "config.h"
#include "pb.h"
#include "pbUser.h"
#include "hash.h"
#include "measure.h"
#include "vectorobj.h"
#include "events.h"
#include <assert.h>
#include <sys/utsname.h>
#include <unistd.h>
#define VN "pb.c: $Revision: 1.27 $ $Date: 2003/01/15 17:49:53 $\n"

#define HASH_SIZE  1025
static Hash*       hashDB_g;
static VectorObj*  routineStack_g;
static FILE*       Fp_out;
static char        fname_g[2048];
static char        nameBuf[2048];
static int         pbInitCalled = 0;
static double      overHeadWall = 0.0;
static double      overHeadCpu  = 0.0;
static double      oHW1,oHW2;
static double      oHC1,oHC2;
static double      junk_sum;
static int         nEvents;
static Events*     events;
static int         warn1 = 1;
static int         warn2 = 1;

void print_record_perl(const char* key, void* adt)
{
  int ievent;
  double cpu,    wall;
  double cpuTtl, wallTtl;
  PBRecord* r = (PBRecord*) adt;
  if (STRCMP(r->key,==,"__/ovrhd")) return;

  cpu     = r->accCpu     - overHeadCpu*r->callCount;
  wall    = r->accWall    - overHeadWall*r->callCount;
  cpuTtl  = r->accCpuTtl  - overHeadCpu*r->callCount;
  wallTtl = r->accWallTtl - overHeadWall*r->callCount;
  if (cpu     < 0.0) cpu     = 0.0;
  if (wall    < 0.0) wall    = 0.0;
  if (cpuTtl  < 0.0) cpuTtl  = 0.0;
  if (wallTtl < 0.0) wallTtl = 0.0;


  fprintf(Fp_out,"$perf{\"%s\"}{calls}    = %lld;\n",    r->key, r->callCount);
  fprintf(Fp_out,"$perf{\"%s\"}{cpu}      = %-15.7g;\n", r->key, cpu);
  fprintf(Fp_out,"$perf{\"%s\"}{wall}     = %-15.7g;\n", r->key, wall);
  fprintf(Fp_out,"$perf{\"%s\"}{comm}     = %lld;\n",    r->key, r->commBytes);
  fprintf(Fp_out,"$perf{\"%s\"}{io}       = %lld;\n",    r->key, r->ioBytes);

  fprintf(Fp_out,"$perf{\"%s\"}{cpu_Ttl}  = %-15.7g;\n", r->key, cpuTtl);
  fprintf(Fp_out,"$perf{\"%s\"}{wall_Ttl} = %-15.7g;\n", r->key, wallTtl);

  for (ievent = 0; ievent < nEvents; ievent++)
    {
      fprintf(Fp_out,"$perf{\"%s\"}{\"%s\"} = %lld;\n",   r->key, 
              events[ievent].pb_name,r->Values(ACC,ievent));
      fprintf(Fp_out,"$perf{\"%s\"}{\"%s%s\"} = %lld;\n",   r->key, 
              events[ievent].pb_name,"_Ttl",r->Values(TtlACC,ievent));
    }
}

void ovrhd_ftn_a(double *xsum)
{
  double a[250], b[250];
  double sum = 0.0;
  double x;
  int i,j;
  static void* pbptr = 0;
  PBstart("__/ovrhd",&pbptr);
  pbptr = 0;

  for (i = 0; i < 250; i++)
    {
      a[i] = 1.0;
      b[i] = 2.0;
    }

  for (j = 0; j < 4; j++)
    for (i = 0; i < 250; i++)
      sum += a[i]*b[i];
  
  *xsum = sum;
  PBend("__/ovrhd");
}

void ovrhd_c_a(double *xsum)
{
  double a[250], b[250];
  double sum = 0.0;
  double x;
  int i,j;
  static void* pbptr = 0;
  PBstart("__/ovrhd",&pbptr);

  for (i = 0; i < 250; i++)
    {
      a[i] = 1.0;
      b[i] = 2.0;
    }

  for (j = 0; j < 4; j++)
    for (i = 0; i < 250; i++)
      sum += a[i]*b[i];
  
  *xsum = sum;
  PBend("__/ovrhd");
}

void ovrhd_b(double *xsum)
{
  double a[250], b[250];
  double sum = 0.0;
  double x;
  int i,j;

  for (i = 0; i < 250; i++)
    {
      a[i] = 1.0;
      b[i] = 2.0;
    }

  for (j = 0; j < 4; j++)
    for (i = 0; i < 250; i++)
      sum += a[i]*b[i];
  
  *xsum = sum;
}

void PB_init(protoFSTRING(fstrfile),   protoFSTRING(fEvents)  
             protoLenFSTRING(fstrfile) protoLenFSTRING(fEvents))
{
  char*  file         = FCD2CP(fstrfile);
  int    len          = FCDLEN(fstrfile);
  char*  eventListStr = FCD2CP(fEvents);
  int    evlen        = FCDLEN(fEvents);
  int    icnt         = 0;
  int    ja, ncalls;
  int    *eventList;
  double startWall, startCpu, endWall, endCpu, sum;
  pbInitCalled = 1;

  if (hashDB_g) {fprintf(stderr,"PB_INIT can only be called once\n"); exit(1);}
  memcpy(&fname_g[icnt],file,len); icnt += len;
  memcpy(&fname_g[icnt],".pb",4);  icnt += 4;
  Fp_out = fopen(&fname_g[0],"w");
  if (Fp_out == 0)
    fprintf(stderr,"Unable to open: %s\n",&fname_g[0]);

  nEvents = parseStr(eventListStr, evlen, &eventList, &events);

  hashDB_g = make_hashtable(HASH_SIZE, hashUC, strcasecmp, strlen);
  VectorObj_Init(routineStack_g);
  INITIALIZE_TIMER;
  counter_init(nEvents, eventList);

  ncalls = 25000;
  START_TIMER(startCpu); START_WALL_TIMER(startWall);
  for (ja = 0; ja < ncalls; ja++)
    ovrhd_ftn_a(&sum);
  END_TIMER(endCpu); END_WALL_TIMER(endWall);
  overHeadWall = endWall - startWall;
  overHeadCpu  = endCpu  - startCpu;
  START_TIMER(startCpu); START_WALL_TIMER(startWall);
  for (ja = 0; ja < ncalls; ja++)
    ovrhd_b(&sum);
  END_TIMER(endCpu); END_WALL_TIMER(endWall);
  overHeadWall -= endWall - startWall;
  overHeadCpu  -= endCpu  - startCpu;
  overHeadWall  = (overHeadWall > 0.0) ? overHeadWall/((double) ncalls) : 0.0;
  overHeadCpu   = (overHeadCpu  > 0.0) ? overHeadCpu /((double) ncalls) : 0.0;
  if (overHeadCpu > overHeadWall)  overHeadCpu = overHeadWall;
  junk_sum = sum;

}

void PB_start(protoFSTRING(fstr) protoLenFSTRING(fstr))
{
  int       ievent;
  char*     fname  = FCD2CP(fstr);
  int       flen   = FCDLEN(fstr);
  INT64     counters[MAX_EVENTS];
  PBRecord* record;
  double    cpu, wall;

  if (!pbInitCalled) return;
  START_TIMER(cpu);
  START_WALL_TIMER(wall);
  counter_read(&counters[0]);

  memcpy(&nameBuf[0],fname,flen);
  nameBuf[flen] = '\0';

  if ((record = find_in_hash(hashDB_g, &nameBuf[0])) == 0)
    {
      char* name;
      PBRecord* result = NULL;
      PBRecord* iret;
    
      ALLOC(name, flen+1, char *);
      memcpy(name,&nameBuf[0], flen+1);
      ALLOC(record, sizeof(PBRecord)+sizeof(INT64)*4*(nEvents-1), PBRecord*);
      record->accCpu      = 0.0;
      record->accWall     = 0.0;
      record->callCount   = 0;
      record->ioBytes     = 0;
      record->commBytes   = 0;
      
      record->accCpuTtl   = 0.0;
      record->accWallTtl  = 0.0;

      for (ievent = 0; ievent < nEvents; ievent++)
        record->Values(ACC,ievent) = record->Values(TtlACC,ievent)  = 0;

      record->key         = name;
      record->len         = flen;
      iret = insert_in_hash(hashDB_g, record->key, record, result);
      assert(iret);
    }

  record->callCount++;
  record->startCpu   = cpu;   record->startCpuTtl   = cpu; 
  record->startWall  = wall;  record->startWallTtl  = wall;
  for (ievent = 0; ievent < nEvents; ievent++)
    record->Values(START,ievent) = record->Values(TtlSTART,ievent) = counters[ievent];

  if (VectorObj_currentSize(routineStack_g) > 0)
    {
      PBRecord* prev  = (PBRecord *) VectorObj_Back(routineStack_g);
      prev->accCpu   += record->startCpu   - prev->startCpu;           
      prev->accWall  += record->startWall  - prev->startWall;          
      for (ievent = 0; ievent < nEvents; ievent++)
        prev->Values(ACC,ievent) += record->Values(START,ievent) - prev->Values(START,ievent);         
    }
  VectorObj_Grow(routineStack_g,record);

}

void PB_end(protoFSTRING(fstr) protoLenFSTRING(fstr))
{
  int       ievent;
  char*     fname  = FCD2CP(fstr);
  int       flen   = FCDLEN(fstr);
  INT64     counters[MAX_EVENTS];
  PBRecord* record;
  int       a,b;

  double endCpu, endWall;


  if (!pbInitCalled) return;
  record = (PBRecord *) VectorObj_Back(routineStack_g);
  #ifndef NDEBUG
    assert(VectorObj_currentSize(routineStack_g) > 0);
    a = (record->len != flen);
    b = STRNCASECMP(fname, !=, record->key, flen);
    if (warn1 && (a || b))
      { 
        warn1 = 0;
        fprintf(stderr,"Warning: PB_START and END do not match, start: %s, end: %.*s\n",
                record->key,flen,fname);
        /* exit(1); */
      }
  #endif

  END_TIMER(endCpu);
  END_WALL_TIMER(endWall);
  counter_read(&counters[0]);

  record->accCpu   += endCpu   - record->startCpu;
  record->accWall  += endWall  - record->startWall;          

  record->accCpuTtl   += endCpu   - record->startCpuTtl;
  record->accWallTtl  += endWall  - record->startWallTtl;          

  for (ievent = 0; ievent < nEvents; ievent++)
    {
      record->Values(ACC,ievent)    += counters[ievent] - record->Values(START,ievent);
      record->Values(TtlACC,ievent) += counters[ievent] - record->Values(TtlSTART,ievent);
    }

  VectorObj_Pop(routineStack_g);
  if (VectorObj_currentSize(routineStack_g) > 0)
    {
      record = (PBRecord *) VectorObj_Back(routineStack_g);
      record->startCpu   = endCpu;
      record->startWall  = endWall;
      for (ievent = 0; ievent < nEvents; ievent++)
        record->Values(START,ievent) = counters[ievent];
    }

}

void PB_report(void)
{
  int            ievent;
  time_t         date;   
  char           dateStr[40];
  struct utsname sysInfo;


  if (!pbInitCalled) return;
  while (VectorObj_currentSize(routineStack_g) > 0)
    {
      PBRecord* record = (PBRecord *) VectorObj_Back(routineStack_g);
      if (warn2)
        {
          fprintf(stderr,"Warning: Failed to call PB_end on routine: %s\n",
                  record->key);
          warn2 = 0;
        }
      VectorObj_Pop(routineStack_g);
    }

  time(&date);
  strftime(dateStr, sizeof(dateStr), "%c\n", localtime(&date));
  uname(&sysInfo);
  fprintf(Fp_out,"*** Time:           %s",   dateStr);
  fprintf(Fp_out,"*** OS:             %s\n", sysInfo.sysname);
  fprintf(Fp_out,"*** HostName:       %s\n", sysInfo.nodename);
  fprintf(Fp_out,"*** OS Release      %s\n", sysInfo.release);
  fprintf(Fp_out,"*** OS Version:     %s\n", sysInfo.version);
  fprintf(Fp_out,"*** Machine:        %s\n", sysInfo.machine);
  fprintf(Fp_out,"*** Username:       %s\n", getlogin());

  
  fprintf(Fp_out,"\n\n### Start Perl Output ###\n");
  fprintf(Fp_out,"$overHeadCpu = %15.7g;\n",overHeadCpu);
  fprintf(Fp_out,"$overHeadWall = %15.7g;\n",overHeadWall);
  fprintf(Fp_out,"@eventNames   = (");
  for (ievent = 0; ievent < nEvents-1; ievent++)
    fprintf(Fp_out,"\"%s\",",events[ievent].pb_name);
  if (ievent < nEvents)
    fprintf(Fp_out,"\"%s\");\n",events[ievent].pb_name);
  else
    fprintf(Fp_out,");\n");

  for_all_in_hash(hashDB_g,print_record_perl);
  fprintf(Fp_out,"### End Perl Output ###\n");
  fclose(Fp_out);
}

void PBreport(void)
{
  PB_report();
}

void PB_commBytes(int* bytes)
{
  PBRecord* record;

  if (!pbInitCalled) return;

  record = (PBRecord *) VectorObj_Back(routineStack_g);
  record->commBytes += (INT64) *bytes;
}

void PB_ioBytes(int* bytes)
{
  PBRecord* record;

  if (!pbInitCalled) return;
  record = (PBRecord *) VectorObj_Back(routineStack_g);
  record->ioBytes += (INT64) *bytes;
}

void PBinit(const char* file, const char* eventListStr)
{
  int    len   = strlen(file);
  int    icnt  = 0;
  int    evlen = strlen(eventListStr);
  int    ja, ncalls;
  int    *eventList;
  double startWall, startCpu, endWall, endCpu, sum;
  pbInitCalled = 1;

  if (hashDB_g) {fprintf(stderr,"PBinit can only be called once\n"); exit(1);}
  memcpy(&fname_g[icnt],file,len); icnt += len;
  memcpy(&fname_g[icnt],".pb",4);  icnt += 4;
  Fp_out = fopen(&fname_g[0],"w");
  if (Fp_out == 0)
    fprintf(stderr,"Unable to open: %s\n",&fname_g[0]);

  nEvents = parseStr(eventListStr, evlen, &eventList, &events);

  hashDB_g = make_hashtable(HASH_SIZE, hashUC, strcasecmp, strlen);
  VectorObj_Init(routineStack_g);
  INITIALIZE_TIMER;
  counter_init(nEvents, eventList);

  ncalls = 25000;
  START_TIMER(startCpu); START_WALL_TIMER(startWall);
  for (ja = 0; ja < ncalls; ja++)
    ovrhd_c_a(&sum);
  END_TIMER(endCpu); END_WALL_TIMER(endWall);
  oHW1 = overHeadWall = endWall - startWall;
  oHC1 = overHeadCpu  = endCpu  - startCpu;
  START_TIMER(startCpu); START_WALL_TIMER(startWall);
  for (ja = 0; ja < ncalls; ja++)
    ovrhd_b(&sum);
  END_TIMER(endCpu); END_WALL_TIMER(endWall);
  /* printf("oHW1: %15.7g, oHC1: %15.7g, oHW2: %15.7g, oHC2: %15.7g\n",oHW1, oHC1,
            endWall - startWall, endCpu  - startCpu); */
  overHeadWall -= endWall - startWall;
  overHeadCpu  -= endCpu  - startCpu;
  overHeadWall  = (overHeadWall > 0.0) ? overHeadWall/((double) ncalls) : 0.0;
  overHeadCpu   = (overHeadCpu  > 0.0) ? overHeadCpu /((double) ncalls) : 0.0;
  if (overHeadCpu > overHeadWall)  overHeadCpu = overHeadWall;
  junk_sum = sum;

}

int PBstart(const char* fname, void *ptr)
{
  int       ievent;
  int       flen   = strlen(fname);
  PBRecord* record = *((PBRecord**) ptr);
  INT64     counters[MAX_EVENTS];
  double cpu, wall;

  assert(pbInitCalled);
  START_TIMER(cpu);
  START_WALL_TIMER(wall);
  counter_read(&counters[0]);

  if (record == 0)
    {
      PBRecord** prec  = ((PBRecord**) ptr);
      memcpy(&nameBuf[0],fname,flen);
      nameBuf[flen] = '\0';

      if ((record = find_in_hash(hashDB_g, &nameBuf[0])) == 0)
        {
          char* name;
          PBRecord* result = NULL;
          PBRecord* iret;
        
          ALLOC(name, flen+1, char *);
          memcpy(name,&nameBuf[0], flen+1);
          ALLOC(record, sizeof(PBRecord)+sizeof(INT64)*4*(nEvents-1), PBRecord*);
          record->accCpu      = 0.0;
          record->accWall     = 0.0;
          record->callCount   = 0;
          record->ioBytes     = 0;
          record->commBytes   = 0;
          
          record->accCpuTtl   = 0.0;
          record->accWallTtl  = 0.0;

          for (ievent = 0; ievent < nEvents; ievent++)
            record->Values(ACC,ievent) = record->Values(TtlACC,ievent)  = 0;

          record->key         = name;
          record->len         = flen;
          iret = insert_in_hash(hashDB_g, record->key, record, result);
          assert(iret);
        }

      *prec = record;
    }
  record->callCount++;
  record->startCpu   = cpu;   record->startCpuTtl   = cpu; 
  record->startWall  = wall;  record->startWallTtl  = wall;
  for (ievent = 0; ievent < nEvents; ievent++)
    record->Values(START,ievent) = record->Values(TtlSTART,ievent) = counters[ievent];

  if (VectorObj_currentSize(routineStack_g) > 0)
    {
      PBRecord* prev  = (PBRecord *) VectorObj_Back(routineStack_g);
      prev->accCpu   += record->startCpu   - prev->startCpu;           
      prev->accWall  += record->startWall  - prev->startWall;          
      for (ievent = 0; ievent < nEvents; ievent++)
        prev->Values(ACC,ievent) += record->Values(START,ievent) - prev->Values(START,ievent);         
    }
  VectorObj_Grow(routineStack_g,record);

  return 0;
}

void PBend(const char* fname)
{
  int       ievent;
  int       flen   = strlen(fname);
  INT64     counters[MAX_EVENTS];
  PBRecord* record;
  int       a,b;

  double endCpu, endWall;


  if (!pbInitCalled) return;
  record = (PBRecord *) VectorObj_Back(routineStack_g);
  #ifndef NDEBUG
    assert(VectorObj_currentSize(routineStack_g) > 0);
    a = (record->len != flen);
    b = STRNCASECMP(fname, !=, record->key, flen);
    if (warn1 && (a || b))
      { 
        warn1 = 0;
        fprintf(stderr,"Warning: PB_START and END do not match, start: %s, end: %.*s\n",
                record->key,flen,fname);
        /* exit(1); */
      }
  #endif

  END_TIMER(endCpu);
  END_WALL_TIMER(endWall);
  counter_read(&counters[0]);

  record->accCpu   += endCpu   - record->startCpu;
  record->accWall  += endWall  - record->startWall;          

  record->accCpuTtl   += endCpu   - record->startCpuTtl;
  record->accWallTtl  += endWall  - record->startWallTtl;          

  for (ievent = 0; ievent < nEvents; ievent++)
    {
      record->Values(ACC,ievent)    += counters[ievent] - record->Values(START,ievent);
      record->Values(TtlACC,ievent) += counters[ievent] - record->Values(TtlSTART,ievent);
    }

  VectorObj_Pop(routineStack_g);
  if (VectorObj_currentSize(routineStack_g) > 0)
    {
      record = (PBRecord *) VectorObj_Back(routineStack_g);
      record->startCpu   = endCpu;
      record->startWall  = endWall;
      for (ievent = 0; ievent < nEvents; ievent++)
        record->Values(START,ievent) = counters[ievent];
    }

}

void PBcommBytes(int bytes)
{
  PBRecord* record;

  if (!pbInitCalled) return;

  record = (PBRecord *) VectorObj_Back(routineStack_g);
  record->commBytes += (INT64) bytes;
}

void PBioBytes(int bytes)
{
  PBRecord* record;

  if (!pbInitCalled) return;
  record = (PBRecord *) VectorObj_Back(routineStack_g);
  record->ioBytes += (INT64) bytes;
}



