/*
 * 
 * Copyright (c) 2001-2002 Robert McLay
 *
 * Authors: Robert McLay <mclay@cfdlab.ae.utexas.edu>
 *          Victor Parr  <vparr@mindspring.com>
 */

/*
 * This file is part of PB: PerfBench.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef _VECTOROBJ_H
#define _VECTOROBJ_H

#include <assert.h>
#include <stdlib.h>

#ifndef MAX
# define MAX(a,b) ((a) > (b)) ? (a) : (b)
#endif

#define VECTOR_INIT_SIZE 512
typedef struct {
  void **buf;
  int   size;      /* allocated size */
  int   current;   /* current size in use */
} VectorObj;

#define VectorObj_Init(v) VectorObj_Init_with_Size(v,VECTOR_INIT_SIZE) 

#define VectorObj_Init_with_Size(v,usrsize) \
   do { v = (VectorObj *) malloc(sizeof(VectorObj));   \
        assert( (v) != 0);                             \
        (v)->size = (usrsize);                         \
        (v)->buf  = (void **)malloc((v)->size*sizeof(void *)); \
        assert( (v)->buf != 0);                        \
        (v)->current = 0;                              \
      } while (0)

#define VectorObj_Free(v) \
   do { free((v)->buf);                                \
        free(v);                                       \
      } while (0)

#define VectorObj_Grow(v, addr)                           \
   do { VectorObj_Resize(v,1);                            \
        (v)->buf[((v)->current)++] = (addr);              \
      } while (0)

#define VectorObj_Pop(v) if ((v)->current > 0) (v)->current--

#define VectorObj_Set(v,i,addr)                                  \
   do { if ( (i) >= (v)->size )                                  \
          VectorObj_ResizeAbsolute((v),(i)+1);                   \
        (v)->buf[(i)] = (addr);                                  \
      } while (0)

#define VectorObj_Finish(v)                                 \
    do { VectorObj_Resize(v,1);                             \
         (v)->buf[((v)->current)++] = 0;                    \
       } while (0)


#define VectorObj_Base(v) (v)->buf

#define VectorObj_End(v) &((v)->buf[(v)->current])

#define VectorObj_Back(v) ((v)->buf[(v)->current-1])

#define VectorObj_Get(v,i) ((v)->buf[(i)])

#define VectorObj_Size(v) ((v)->size)

#define VectorObj_currentSize(v) ((v)->current)

#define VectorObj_Resize(v,len) \
   do { if ((len)  > ((v)->size - (v)->current))                   \
          {                                                        \
            (v)->size = MAX( 2*((v)->size), (v)->current + (len)); \
            (v)->buf  = (void **)realloc((v)->buf,(v)->size*sizeof(void*)); \
            assert( (v)->buf != 0);                                \
          }                                                        \
      } while (0)

#define VectorObj_ResizeAbsolute(v,len)                            \
   do { if ((len)  > (v)->size)                                    \
          {                                                        \
            (v)->size = MAX( 2*((v)->size), (len));                \
            (v)->buf  = (void **)realloc((v)->buf,(v)->size*sizeof(void*)); \
            assert( (v)->buf != 0);                                \
          }                                                        \
      } while (0)



#define VectorObj_Reset(v)  (v)->current = 0

#endif


